<?php
set_error_handler(
  function( $errno, $errstr, $errfile, $errline ) {
    throw new ErrorException( $errstr, $errno, 0, $errfile, $errline ) ;
  }
) ;

set_exception_handler(function($e) {
  echo "$e\n" ;
  debug() ;
}) ;

// echo title
function echoTitle( $data_udp, $localhost, $listenPort, $datafile ) {
  global $PGNAME, $VERSION ;

  printf( "-----------------------------\n" ) ;
  printf( "%s %s\n", $PGNAME, $VERSION ) ;
  printf( "-----------------------------\n" ) ;
  if( $data_udp ) {
    printf( "listening on %s:%s\n", $localhost, $listenPort ) ;
    printf( "ready to go\n" ) ;
  } else {
    printf( "reading from file: %s\n", $datafile ) ;
  }
}

// init world data master
function initWDataMst() {
  global $CHECKPOINT_NUM, $LAPTIME_MAX ;

  $wDataMst = array(
    "m_time" => $LAPTIME_MAX,
    "m_raceTime" => 0 ,
    "m_sessionType" => -1,
    "m_track_number" => -1,
    "m_num_cars" => 0,
    "m_interval_count" => 0,
    "m_checkPointDistance" => 200,
    "m_bestLapTime" => $LAPTIME_MAX,
    "m_bestLapTimeTID" => 99 ,
    "m_bestLapTimeDN" => "",
    "m_bestLapTimeDNS" => "",
    "m_bestLapTimeTC" => 0,
    "m_bestLapTimeL" => 0,
    "m_bestSector1Time" => $LAPTIME_MAX,
    "m_bestSector1TimeTID" => 99 ,
    "m_bestSector1TimeDN" => "",
    "m_bestSector1TimeDNS" => "",
    "m_bestSector1TimeTC" => 0,
    "m_bestSector1TimeL" => 0,
    "m_bestSector2Time" => $LAPTIME_MAX,
    "m_bestSector2TimeTID" => 99 ,
    "m_bestSector2TimeDN" => "",
    "m_bestSector2TimeDNS" => "",
    "m_bestSector2TimeTC" => 0,
    "m_bestSector2TimeL" => 0,
    "m_bestSector3Time" => $LAPTIME_MAX,
    "m_bestSector3TimeTID" => 99 ,
    "m_bestSector3TimeDN" => "",
    "m_bestSector3TimeDNS" => "",
    "m_bestSector3TimeTC" => 0,
    "m_bestSector3TimeL" => 0,
    "m_fuelLeft" => 0 ,
    "m_fuelUse" => array( 0, 0, 0, 0 ),
    "m_fuelDistance" => array( 0, 0, 0, 0 ),
    "m_prevDistance" => 0 ,
  ) ;
  return $wDataMst ;
}

// init car data master
function initCDataMst( $idx, $wData, $cd1, &$wDataMst ) {
  global $LAPTIME_MAX, $CARS_MAX, $CHECKPOINT_NUM, $TYRES ;

  $cdm1 = array(
    "m_lastLapTime" => $LAPTIME_MAX,
    "m_bestLapTime" => $LAPTIME_MAX,
    "m_currentLapNum" => 0,
    "m_sector" => 0,
    "m_penalties" => 0,
    "m_inPits" => 0,
    "m_inPitTime" => 0,
    "m_pitStops" => 0,
    "m_pitLaps" => 0,
    "m_driverName" => "",
    "m_driverNameShort" => "",
    "m_teamName" => "",
    "m_bestSector1Time" => $LAPTIME_MAX,
    "m_bestSector2Time" => $LAPTIME_MAX,
    "m_bestSector3Time" => $LAPTIME_MAX,
    "m_bestSector1TimeSave" => $LAPTIME_MAX,
    "m_bestSector2TimeSave" => $LAPTIME_MAX,
    "m_bestSector3TimeSave" => $LAPTIME_MAX,
    "m_carPositionHigh" => $CARS_MAX,
    "m_carPositionLow" => 1,
    "m_lapTime" => array(),
    "m_sector1Time" => array(),
    "m_sector2Time" => array(),
    "m_sector3Time" => array(),
    "m_carPosition" => array(),
    "m_tyreCompound" => array(),
    "m_totalDistance" => 0,
    "m_checkPointCount" => 0,
    "m_checkPointTime" => array(),
    "m_currentLapInvalid" => array(),
    "m_interval" => 0,
    "m_speedTrapTime1" => 0,
    "m_speedTrapTime2" => 0,
    "m_speedTrapDist1" => 0,
    "m_speedTrapDist2" => 0,
    "m_speedTrapKPH" => 0,
  ) ;

  $cln = $cd1["m_currentLapNum"] ;
  $cp  = $cd1["m_carPosition"] ;
  $blt = $cd1["m_bestLapTime"] ;

  // set driver name and initial position
  setDriverName( $idx, $wData["m_player_car_index"], $wData["m_era"], $cd1, $cdm1 ) ;

  // set default data for race already started
  $cdm1["m_carPosition"][] = $cp ;
  for( $i=0 ; $i<($cln-1) ; $i++ ) {
    $cdm1["m_currentLapInvalid"][$i] = 0 ;
    $cdm1["m_sector1Time"][$i] = 0 ;
    $cdm1["m_sector2Time"][$i] = 0 ;
    $cdm1["m_sector3Time"][$i] = 0 ;
    $cdm1["m_lapTime"][$i] = 0 ;
    $cdm1["m_carPosition"][$i] = $cp ;
  }

  // set session best
  if( $blt > 0 && $blt < $wDataMst["m_bestLapTime"] ) {
    $key_blt = "m_bestLapTime" ;
    $wDataMst[$key_blt] = $blt ;
    $wDataMst[$key_blt . "TID"] = $cd1["m_teamId"] ;
    $wDataMst[$key_blt . "DN"] = $cdm1["m_driverName"] ;
    $wDataMst[$key_blt . "DNS"] = $cdm1["m_driverNameShort"] ;
    $wDataMst[$key_blt . "TC"] = sizeof( $TYRES ) - 1 ;
    $wDataMst[$key_blt . "L"] = 0 ;
  }

  return $cdm1 ;
}

function initIntervalData( $wData, $cData, $wDataMst, &$cDataMst ) {
  global $CHECKPOINT_NUM ;
 
  $rpos = makeRacePositionData( $wData, $cData ) ;
  $cpnum = $CHECKPOINT_NUM ;
  $cpcsv = $cpnum ;
  $ts = $wData["m_track_size"] ;
  foreach( $rpos as $i => $v ) {
    $cd1 = $cData[$i] ;
    $cdm1 = &$cDataMst[$i] ;
    $cln = $cd1["m_currentLapNum"] ;
    $ld = $cd1["m_lapDistance"] ;
    $cpc = floor( ($ld / $wDataMst["m_checkPointDistance"]) + 
                  ($cpnum * ($cln-1) ) ) ;
//printf( "%d,%s,%d,%.2f\n", $v, $cdm1["m_driverNameShort"], $cln, $ld ) ;
    $chk = $ld / $ts ;

    if( $chk > 0.25 || $cpc > $cpcsv || $cpc < 0 ) {
      break ;
    }
    for( $j=1 ; $j<=$cpc ; $j++ ) {
      $cdm1["m_checkPointTime"][$j] = 0 ;
    }
    $cdm1["m_checkPointCount"] = $cpc ;
    $cpcsv = $cpc ;
  }

  return ;
}

// get world data from D-BOX
function getWData( $ud ) {
  $wData = array(
    "m_time" => $ud["d01_1"],
    "m_totalDistance" => $ud["d01_4"],
    "m_drs" => $ud["d01_43"],
    "m_fuel_in_tank" => $ud["d01_46"],
    "m_fuel_capacity" => $ud["d01_47"],
    "m_brakes_temp" => array(
      $ud["d01_54"], $ud["d01_55"], $ud["d01_52"], $ud["d01_53"],
    ),
    "m_tyres_pressure" => array(
      $ud["d01_58"], $ud["d01_59"], $ud["d01_56"], $ud["d01_57"],
    ),
    "m_total_laps" => $ud["d01_61"],
    "m_track_size" => $ud["d01_62"],
    "m_sessionType" => $ud["d01_67"],
    "m_drsAllowed" => $ud["d01_68"],
    "m_track_number" => $ud["d01_69"],
    "m_vehicleFIAFlags" => $ud["d01_70"],
    "m_era" => $ud["d01_71"],
    "m_engine_temperature" => $ud["d01_72"],
    "m_tyres_temperature" => array(
      $ud["d02_3"], $ud["d02_4"], $ud["d02_1"], $ud["d02_2"],
    ),
    "m_tyres_wear" => array(
      $ud["d02_7"], $ud["d02_8"], $ud["d02_5"], $ud["d02_6"],
    ),
    "m_front_brake_bias" => $ud["d02_10"],
    "m_fuel_mix" => $ud["d02_11"],
    "m_tyres_damage" => array(
      $ud["d02_15"], $ud["d02_16"], $ud["d02_13"], $ud["d02_14"],
    ),
    "m_front_left_wing_damage" => $ud["d02_17"],
    "m_front_right_wing_damage" => $ud["d02_18"],
    "m_rear_wing_damage" => $ud["d02_19"],
    "m_engine_damage" => $ud["d02_20"],
    "m_gear_box_damage" => $ud["d02_21"],
    "m_exhaust_damage" => $ud["d02_22"],
    "m_pit_speed_limit" => $ud["d02_24"],
    "m_session_time_left" => $ud["d03_1"],
    "m_num_cars" => $ud["d04_4"],
    "m_player_car_index" => $ud["d04_5"],
  ) ;

  return $wData ;
}

// get cars data from D-BOX
function getCData( $ud, $m_num_cars ) {
  global $CARS_MAX ;

  $cData = array() ;
  for( $i=0 ; $i<$CARS_MAX ; $i++ ) {
    $key1 = sprintf("d%da_",$i+11 ) ;
    $key2 = sprintf("d%db_",$i+11 ) ;
    $cData[$i] = array(
      "m_lastLapTime" => $ud["${key1}4"],
      "m_currentLapTime" => $ud["${key1}5"],
      "m_bestLapTime" => $ud["${key1}6"],
      "m_sector1Time" => $ud["${key1}7"],
      "m_sector2Time" => $ud["${key1}8"],
      "m_lapDistance" => $ud["${key1}9"],
      "m_driverId" => $ud["${key2}1"],
      "m_teamId" => $ud["${key2}2"],
      "m_carPosition" => $ud["${key2}3"],
      "m_currentLapNum" => $ud["${key2}4"],
      "m_tyreCompound" => $ud["${key2}5"],
      "m_inPits" => $ud["${key2}6"],
      "m_sector" => $ud["${key2}7"],
      "m_currentLapInvalid" => $ud["${key2}8"],
      "m_penalties" => $ud["${key2}9"],
    ) ;
  }
  return $cData ;
}

// reload config
function reloadConfig( $wData, $cData, &$cDataMst ) {
  global $CARS_MAX ;

  include "relay.php" ;
  include "drivers_custom.php" ;
  for( $i=0 ; $i<$CARS_MAX ; $i++ ) {
    $cd1 = $cData[$i] ;
    $cdm1 = &$cDataMst[$i] ;
    setDriverName( $i, $wData["m_player_car_index"], $wData["m_era"], $cd1, $cdm1 ) ;
  }
}

// start session
function startSession( $wData, $cData, &$wDataMst, &$cDataMst ) {
  global $DRIVERS, $CARS_MAX, $CHECKPOINT_NUM ;

  // initialize
  $nc = $wData["m_num_cars"] ;
  $cpnum = $CHECKPOINT_NUM ;
  $wDataMst = initWDataMst() ;
  $wDataMst["m_time"] = $wData["m_time"] ;
  $wDataMst["m_track_number"] = $wData["m_track_number"] ;
  $wDataMst["m_num_cars"] = $nc ;
  $wDataMst["m_sessionType"] = $wData["m_sessionType"] ;
  $wDataMst["m_checkPointDistance"] = $wData["m_track_size"] / $cpnum ;
  $wDataMst["m_fuelLeft"] = $wData["m_fuel_in_tank"] ;

  $cDataMst = array() ;
  for( $i=0 ; $i<$CARS_MAX ; $i++ ) {
    $cd1 = $cData[$i] ;
    $cDataMst[$i] = initCDataMst( $i, $wData, $cd1, $wDataMst ) ;
  }
  // initialize interval data
  initIntervalData( $wData, $cData, $wDataMst, $cDataMst ) ;
}

// set driver name
function setDriverName( $i, $pci, $era, $cd1, &$cdm1 ) {
  global $DRIVERS, $DRIVERS_CUSTOM, $USE_DRIVERS_CUSTOM, $PLAYER ;

  if( $i == $pci ) {
    // player
    $sz = sizeof( $PLAYER ) ;
    if( $sz == 2 && strlen($PLAYER[0]) > 0 ) {
      $cdm1["m_driverName"] = $PLAYER[0] ;
      $cdm1["m_driverNameShort"] = $PLAYER[1] ;
    } else {
      $cdm1["m_driverName"] = "Player" ;
      $cdm1["m_driverNameShort"] = "YOU" ;
    }
  } else {
    // others
    $did = $cd1["m_driverId"] ;
    if( $USE_DRIVERS_CUSTOM  == 1 ) {
      // custom drivers
      $l1 = getArray( $DRIVERS_CUSTOM, $era, array() ) ;
      $dmst = getArray( $l1, $did, array() ) ;
    } else {
      // official drivers
      $dmst = $DRIVERS[$era][$did] ;
    }
    $cdm1["m_driverName"] = getArray( $dmst, 0, "Driver Unknown" ) ;
    $cdm1["m_driverNameShort"] = getArray( $dmst, 1, "ERR" ) ;
  }
}

// update status
function updateStatus( $shmid, $wData, $cData, &$wDataMst, &$cDataMst ) {
  global $CARS_MAX, $INTERVAL_TIME, $INTERVAL_DIST, $CHECKPOINT_NUM, $SPDMAX, $SPDTRAP_DIST, $SPDTRAP_MULTI ;

  $rt = $wData["m_time"] ; // race time
  $st = $wData["m_sessionType"] ; // sessionType
  $nc = $wData["m_num_cars"] ;
  $ts = $wData["m_track_size"] ; // track size(length)
  $tn = $wData["m_track_number"] ; // track number
  $tln = $wData["m_total_laps"] ;
  $cpd = $wDataMst["m_checkPointDistance"] ; // track size / CHECKPOINT_NUM
  $pci = $wData["m_player_car_index"] ;
  $rdist = $tln * $ts ; // race distance
  $rpos = array() ;
  $fom = array() ;
  $wDataMst["m_time"] = $rt ;
//echo "$st," . $cData[$pci]["m_lapDistance"] . "\n";

  // make race position data
  $rpos = makeRacePositionData( $wData, $cData ) ;
  // make track position data
  $lidx = 0 ;
  foreach( $rpos as $i => $v ) {
    $lidx = $i ;
    break ;
  }
  $tpos = makeTrackPositionData( $st, $ts, $lidx, $wData, $cData, $cDataMst ) ;

  // do each car
  for( $i=0 ; $i<$CARS_MAX ; $i++ ) {
    $cd1 = $cData[$i] ; // 1 car data
    $cdm1 = &$cDataMst[$i] ; // 1 car data master
    $clt = $cd1["m_currentLapTime"] ;
    $cln = $cd1["m_currentLapNum"] ; // current lap 
    $ld = $cd1["m_lapDistance"] ; // current lap distance
    $dist = $ts * ($cln-1) + $ld ; // total distance

    // record invalid lap
    $key = "m_currentLapInvalid" ;
    $cdm1[$key][$cln-1] = $cd1[$key] ;

    // calc fuel data(player only)
    if( $i == $pci ) {
      if( $cd1["m_inPits"] == 0 ) {
        checkFuel( $st, $ld, $ts, $clt, $wData, $wDataMst ) ;
      }
    }
    // make interval data
    if( $st != 3 ||
        ($st == 3 && $clt > 0 && $dist <= ($rdist + $cpd * 2)) ) {
      // record time at the check point
      $cpc = floor( $dist / $cpd) ;
      if( $cpc > 0 ) {
        $dist_s = $cpc * $cpd ;
        $dist_e = $dist_s + 10 ;
        if( $dist_s <= $dist && $dist < $dist_e ) {
          $cdm1["m_checkPointTime"][$cpc] = $rt ;
          $cdm1["m_checkPointCount"] = $cpc ;
        }
      }
      // check pit
      checkPit( $st, $cln, $ld, $wData, $cd1, $wDataMst, $cdm1 ) ;
      // check speed trap
      checkSpeedTrap( $cln, $tn, $clt, $ld, $cdm1 ) ;
      // record sector time
      $ret = recordSectorTime( $st, $cln, $cd1, $wDataMst, $cdm1 ) ;
      // restore session best
      if( $ret != 0 ) {
        restoreSessionBest( $wData, $wDataMst, $cDataMst ) ;
      }
    }
  }

  // check interval
  checkInterval( $rpos, $wData, $cData, $cDataMst ) ;

  if( $cData[$pci]["m_currentLapTime"] > 0 ) {
    outputHtml( $shmid, $rpos, $tpos, $wData, $cData, $wDataMst, $cDataMst ) ;
  }
}

// make race position data
function makeRacePositionData( $wData, $cData ) {
  global $CARS_MAX ;

  $rpos = array() ;
  for( $i=0 ;$i<$CARS_MAX ; $i++ ) {
    $cp = $cData[$i]["m_carPosition"] ;
    if( $cp > 0 ) {
      $rpos[$i] = $cp ;
    }
  }
  asort( $rpos ) ;

  return $rpos ;
}

// make track position data
function makeTrackPositionData( $st, $ts, $lidx, $wData, $cData, $cDataMst ) {
  global $CARS_MAX ;

  // make array index=>lapDistance
  $tmp = array() ;
  $tpos = array() ;
  for( $i=0 ;$i<$CARS_MAX ; $i++ ) {
    $cd1 = $cData[$i] ;
    $cdm1 = $cDataMst[$i] ;
    $in_pits = ($cd1["m_inPits"] != 0) ;
    $cln = $cd1["m_currentLapNum"] ;
    $ld = $cd1["m_lapDistance"] ;
    $cp = $cd1["m_carPosition"] ;
    $dist = $ts * ($cln-1) + $ld ; // total distance
    if( $cp > 0 ) {
      if( ($st == 3 && !checkRetire( $st, $wData, $cd1, $cdm1 )) ||
          ($st != 3 && $in_pits == 0 && $dist > ($ts*0.06)) ) {
        $tmp[$i] = $ld ;
      }
    }
  }
  // sort descending lapDistance
  arsort( $tmp ) ;
  $sz = sizeof( $tmp ) ;

  // make array from leader to bottom
  $out = FALSE ;
  foreach( $tmp as $i => $ld ) {
    if( $i == $lidx ) {
      $out = TRUE ;
    }
    if( $out ) {
      $tpos[$i] = $ld ;
    }
  }

  // add array until $CARS_MAX
  foreach( $tmp as $i=> $ld ) {
    if( sizeof($tpos) >= $sz ) {
      break ;
    } else {
      $tpos[$i] = $ld ;
    }
  }
  
  return $tpos ;
}
// check fuel data
function checkFuel( $st, $ld, $ts, $clt, $wData, &$wDataMst ) {
  global $FUEL_MIX_MULTI ;

  $td = $wData["m_totalDistance"] ;
  $fit = $wData["m_fuel_in_tank"] ;
  $fm = $wData["m_fuel_mix"] ;
  $fl = &$wDataMst["m_fuelLeft"] ;
  $fu = &$wDataMst["m_fuelUse"] ;
  $futmp = 0 ;

  while( $td < 0 ) {
    $td += $ts ;
  }
  if( $clt > 0 ) {
    $futmp = ($fl - $fit) ;
  }

  $rd = $td - $wDataMst["m_prevDistance"] ;
  if( $rd > 0 && $rd < 50 && $futmp > 0 ) {
    $fu[$fm] += $futmp ;
    $wDataMst["m_fuelDistance"][$fm] += $rd ;
    $fl = $fit ;
    $wDataMst["m_prevDistance"] = $td ;
  }
  if( $rd > 50 || $rd < 0 ) {
    $fl = $fit ;
    $wDataMst["m_prevDistance"] = $td ;
  }
}

// check pit stops
function checkPit( $st, $cln, $ld, $wData, $cd1, &$wDataMst, &$cdm1 ) {
  $pit = $cd1["m_inPits"] ;
  $pitm = $cdm1["m_inPits"] ;

  // set inPits status and increment pit stops
  if( $pit != 0 ) {
    if( $pitm == 0 ) {
      $cdm1["m_inPitTime"] = $wData["m_time"] ;
      $cdm1["m_inPits"] = $pit ;
      $cdm1["m_pitStops"]++ ;
    }
  }
  // clear inPits status and record last stop laps
  if( $pit == 0 && $pitm != 0 ) {
    $cdm1["m_inPits"] = 0 ;
    $cdm1["m_pitLaps"] = $cd1["m_currentLapNum"] ;
    if( $st != 3 ) {
      // clear old sector time for practice and qualifying
      $cdm1["m_sector1Time"][$cln-1] = 0 ;
      $cdm1["m_sector2Time"][$cln-1] = 0 ;
      // clear fuel data for practice and qualifying
      $wDataMst["m_fuelLeft"] = $wData["m_fuel_in_tank"] ;
      $wDataMst["m_fuelUse"] = array( 0, 0, 0, 0 ) ;
      $wDataMst["m_fuelDistance"] = array( 0, 0, 0, 0 ) ;
      $wDataMst["m_prevDistance"] = $wData["m_totalDistance"] ;
    }
  }
}

// record sector time
function recordSectorTime( $st, $cln, $cd1, &$wDataMst, &$cdm1 ) {
  // sector 1
  recordSectorTimeSub1( $cln, "m_sector1Time", "m_bestSector1Time", $cd1, $wDataMst, $cdm1 ) ;
  // sector 2
  recordSectorTimeSub1( $cln, "m_sector2Time", "m_bestSector2Time", $cd1, $wDataMst, $cdm1 ) ;
  // sector 3
  return recordSectorTimeSub3( $st, $cln, "m_sector3Time", "m_bestSector3Time", $cd1, $wDataMst, $cdm1 ) ;
}

// record sector time for sector 1&2
function recordSectorTimeSub1( $cln, $key, $keyBest, $cd1, &$wDataMst, &$cdm1 ) {
  $t = $cd1[$key] ;
  $tm = &$cdm1[$key] ;

  if( $t > 0 ) {
/*
if(strcmp($cdm1["m_driverNameShort"],"NEK")==0){
echo "$cln,$t\n";
}
*/
    $tm[$cln-1] = $t ;
    $cli = getArray( $cdm1["m_currentLapInvalid"], $cln-1 ) ;
    if( $cli == 0 ) {
      // sector personal best
      if( $t < $cdm1[$keyBest] ) {
        $cdm1[$keyBest] = $t ;
      }
      // sector session best
      if( $t < $wDataMst[$keyBest] ) {
        $wDataMst[$keyBest] = $t ;
        $wDataMst[$keyBest . "TID"] = $cd1["m_teamId"] ;
        $wDataMst[$keyBest . "DN"] = $cdm1["m_driverName"] ;
        $wDataMst[$keyBest . "DNS"] = $cdm1["m_driverNameShort"] ;
        $wDataMst[$keyBest . "TC"] = $cd1["m_tyreCompound"] ;
        $wDataMst[$keyBest . "L"] = $cln ;
      }
    }
  }
}

// record sector time for sector 3
function recordSectorTimeSub3( $st, $cln, $key, $keyBest, $cd1, &$wDataMst, &$cdm1 ) {
  global $LAPTIME_MAX, $SECTORTIME_MAX ;

  $ret = 0 ;
  $tm = &$cdm1[$key] ;
  $ld = $cd1["m_lapDistance"] ;
  $llt = $cd1["m_lastLapTime"] ;
  if( $cln > (sizeof($tm)+1) && $llt > 0 && $ld > 15 ) {
    $cln2 = $cln-2 ;

    // record data at control line
    $wDataMst["m_raceTime"] += $llt ;
    $cdm1["m_lapTime"][$cln2] = $llt ;
    $cdm1["m_carPosition"][$cln-1] = $cd1["m_carPosition"] ;

    // record sector 3 time
    $s1t = getArray( $cdm1["m_sector1Time"], $cln2, $SECTORTIME_MAX ) ;
    $s2t = getArray( $cdm1["m_sector2Time"], $cln2, $SECTORTIME_MAX ) ;
    $s12t = $s1t + $s2t ;

    $s3t = $LAPTIME_MAX ;
    if( $s1t > 0 && $s2t > 0 && $llt > $s12t ) {
      $s3t = $llt - $s12t ;
    }
    $tm[$cln2] = $s3t ;

    $cli2 = getArray( $cdm1["m_currentLapInvalid"], $cln2 ) ;
    if( $cli2 == 0 ) {
      // lap valid
      // sector personal best
      if( $s3t < $cdm1[$keyBest] ) {
        $cdm1[$keyBest] = $s3t ;
      }
      // sector session best
      if( $s3t < $wDataMst[$keyBest] ) {
        $wDataMst[$keyBest] = $s3t ;
        $wDataMst[$keyBest . "TID"] = $cd1["m_teamId"] ;
        $wDataMst[$keyBest . "DN"] = $cdm1["m_driverName"] ;
        $wDataMst[$keyBest . "DNS"] = $cdm1["m_driverNameShort"] ;
        $wDataMst[$keyBest . "TC"] = $cd1["m_tyreCompound"] ;
        $wDataMst[$keyBest . "L"] = $cln-1 ;
      }

      if( $llt > 0 ) {
        $key_blt = "m_bestLapTime" ;
        // lap personal best
        if( $llt < $cdm1[$key_blt] ) {
          $cdm1[$key_blt] = $llt ;
        }
        // lap session best
        if( $llt < $wDataMst[$key_blt] ) {
          $wDataMst[$key_blt] = $llt ;
          $wDataMst[$key_blt . "TID"] = $cd1["m_teamId"] ;
          $wDataMst[$key_blt . "DN"] = $cdm1["m_driverName"] ;
          $wDataMst[$key_blt . "DNS"] = $cdm1["m_driverNameShort"] ;
          $wDataMst[$key_blt . "TC"] = $cd1["m_tyreCompound"] ;
          $wDataMst[$key_blt . "L"] = $cln-1 ;
        }
      }
    } else {
      // lap invalid
      // restore valid personal best
      $cdm1["m_bestSector1Time"] = $cdm1["m_bestSector1TimeSave"] ;
      $cdm1["m_bestSector2Time"] = $cdm1["m_bestSector2TimeSave"] ;
      $cdm1["m_bestSector3Time"] = $cdm1["m_bestSector3TimeSave"] ;
      // restore valid session best
      $ret = 1 ;
    }
    // backup best time
    $cdm1["m_bestSector1TimeSave"] = $cdm1["m_bestSector1Time"] ;
    $cdm1["m_bestSector2TimeSave"] = $cdm1["m_bestSector2Time"] ;
    $cdm1["m_bestSector3TimeSave"] = $cdm1["m_bestSector3Time"] ;

/*
echo "L$cln-cdm1 " ;
print_r($cdm1) ;
*/
  }
  return $ret ;
}

// restore session best
function restoreSessionBest( $wData, &$wDataMst, $cDataMst ) {
  for( $i=1 ; $i<=3 ; $i++ ) {
    $time = array() ;
    $key = sprintf( "m_bestSector%dTime", $i ) ;
    for( $j=0 ; $j<$wData["m_num_cars"] ; $j++ ) {
      $time[] = $cDataMst[$j][$key] ;
    }
    sort( $time ) ;
    $wDataMst[$key] = $time[0] ;
  }
}

// check interval
function checkInterval( $rpos, $wData, $cData, &$cDataMst ) {
  global $LAPTIME_MAX ;
  $time = $wData["m_time"] ;
  $st = $wData["m_sessionType"] ;

  // get data for first position
  foreach( $rpos as $i => $v ) {
    $cdf = $cData[$i] ;
    $cdmf = $cDataMst[$i] ;
    break ;
  }

  switch( $st ) {
    case 1: // practice
    case 2: // qualifying
      // calculate interval against a leader
      $bltl = $cdf["m_bestLapTime"] ; 
      foreach( $rpos as $i => $v ) {
        $cdt = $cData[$i] ;
        $cdmt = &$cDataMst[$i] ;
        $bltt = $cdt["m_bestLapTime"] ; 
        if( $bltt != 0 && $bltt != $LAPTIME_MAX ) {
          $cdmt["m_interval"] = $bltt - $bltl ;
        }
      }
      break ;
    case 3: // race
      // calculate interval for cars befor and after
      foreach( $rpos as $i => $v ) {
        $cdmt = &$cDataMst[$i] ;
        $cpc = $cdmt["m_checkPointCount"] ;
        if( $cpc > 0 ) {
          $bcpt_t = array_key_exists( "$cpc", $cdmt["m_checkPointTime"] ) ;
          $bcpt_f = array_key_exists( "$cpc", $cdmf["m_checkPointTime"] ) ;
          if( $bcpt_t && $bcpt_f ) {
            $cptt = getArray( $cdmt["m_checkPointTime"], $cpc ) ;
            $cptf = getArray( $cdmf["m_checkPointTime"], $cpc ) ;
            $cdmt["m_interval"] = $cptt - $cptf ;
          }
        }
        $cdmf = $cDataMst[$i] ;
      }
      break ;
  }
}

// check speed trap
function checkSpeedTrap( $cln, $tn, $clt, $ld, &$cdm1 ) {
  global $SPDTRAP_DIST ;
  $SPDTRAP_LENGTH = 33 ;
  $SPDTRAP_RANGE = 3 ;

  $spdtd = $SPDTRAP_DIST[$tn] ;
  $spdtd1 = $spdtd -  $SPDTRAP_LENGTH ;
  $spdtd2 = $spdtd - $SPDTRAP_LENGTH * 2 ;
  // speed trap1
  if( $spdtd1 <= $ld && $ld <= $spdtd1+$SPDTRAP_RANGE ) { 
    $cdm1["m_speedTrapTime1"] = $clt ;
    $cdm1["m_speedTrapDist1"] = $ld ;
  }
  // speed trap2
  if( $spdtd2 <= $ld && $ld <= $spdtd2+$SPDTRAP_RANGE ) { 
    $cdm1["m_speedTrapTime2"] = $clt ;
    $cdm1["m_speedTrapDist2"] = $ld ;
  }
  if( $spdtd <= $ld && $ld <= $spdtd+$SPDTRAP_RANGE ) { 
    $spdtt1 = getArray( $cdm1, "m_speedTrapTime1" ) ;
    $spdtl1 = getArray( $cdm1, "m_speedTrapDist1" ) ;
    $spdtt2 = getArray( $cdm1, "m_speedTrapTime2" ) ;
    $spdtl2 = getArray( $cdm1, "m_speedTrapDist2" ) ;
    if( $spdtt1 > 0 && $spdtt1 > $spdtt2 && $spdtl1 > 0 && $spdtl1 > $spdtl2 ) {
      $spdavg = ($ld - $spdtl2) / ($clt - $spdtt2) * 3.6 ;
      $spdmax = ($ld - $spdtl1) / ($clt - $spdtt1) * 3.6 ;
      $spdmin = ($spdtl1 - $spdtl2) / ($spdtt1 - $spdtt2)* 3.6  ;
      $kph = $spdmax + 
        ( ( ($spdmax - $spdavg) + ($spdavg - $spdmin) ) / 2 ) ;
      $cdm1["m_speedTrapTime1"] = 0 ;
      $cdm1["m_speedTrapDist1"] = 0 ;
      $cdm1["m_speedTrapTime2"] = 0 ;
      $cdm1["m_speedTrapDist2"] = 0 ;
      if( $cdm1["m_speedTrapKPH"] < $kph ) {
        $cdm1["m_speedTrapKPH"] = $kph ;
      }
    }
  }
}

// get sector time
function getSectorTime( $cdm1, $cln, $key ) {
  $idx = $cln - 1 ;
  if( $cln > 0 ) {
    $tm = getArray( $cdm1[$key], $idx ) ;
  } else {
    $tm = 0 ;
  }
  return $tm ;
}

// output formatted time
function outputTime( $time, $flag ) {
  global $LAPTIME_MAX ;

  $tmp = $time ;
  $h = floor( $time / 3600 ) ;
  $tmp -= ($h * 3600) ;
  $m = floor( $tmp / 60 ) ;
  $tmp -= ($m * 60) ;
  $s = floor( $tmp ) ;
  $tmp -= $s ;
  $ms = floor( $tmp * 1000 ) ;

  if( $time > 0 && $time != $LAPTIME_MAX ) {
    if( $time < 60 ) {
      $t = sprintf("%2d.%03d", $s, $ms ) ;
    } else {
      if( $time > 3600 ) {
        $t = sprintf("%d:%02d:%02d.%03d", $h, $m, $s, $ms ) ;
      } else {
        $t = sprintf("%d:%02d.%03d", $m, $s, $ms ) ;
      }
    }
    switch( $flag ) {
      case 1:
        printf("+%s", $t ) ;
        break ;
      default:
        printf("%s", $t ) ;
        break ;
    }
  }
}

// check best time
function checkBestTime( $time, $key, $wDataMst, $cdm1 ) {
  $class = "none" ;
  if( $time != 0 ) {
    if( $time <= $cdm1[$key] ) {
      $class = "pb" ;
    }
    if( $time <= $wDataMst[$key] ) {
      $class = "sb" ;
    }
  }
  return $class ;
}

/* check retire */
function checkRetire( $st, $wData, $cd1, $cdm1 ) {
  global $RETIRE_SEC ;

  $pit = $cd1["m_inPits"] ;
  $pittime = $wData["m_time"] - $cdm1["m_inPitTime"] ;
  if( $st == 3 && $pit != 0 && $pittime >= $RETIRE_SEC ) {
    $ret = TRUE ;
  } else {
    $ret = FALSE ;
  }

  return $ret ;
}

/* get array data safely */
function getArray( $arr, $key, $default=0 ) {
  if( array_key_exists("$key", $arr) ) {
    $ret = $arr[$key] ;
  } else {
    $ret = $default ;
  }
  return $ret ;
}

function getRaceTime( $laptimes ) {
  $ret = 0 ;
  foreach( $laptimes as $time ) {
    $ret += $time ;
  }
  return $ret ;
}

// speed indicator
function makeSpeedIndicator( $spd ) {
  $spd_min = 180 ;
  $spd_max = 350 ;
  $cb_max = 255 ;
  $cg_max = 160 ;
  $cr_max = 160 ;

  $res = 255 + 255 ;
  $p1 = $res / 2 ;
  $p2 = $res * 3 / 4 ;
  $p3 = $p2 - $res / 12 ;
  $unit = ($spd_max - $spd_min + 1 ) / $res ;

  $uidx = ($spd - $spd_min) / $unit ;
  if( $p1 >= $uidx ) {
    $cr = 255 - $uidx ;
    $cg = 255 ;
    $cb = 255 ;
  }
  if( $p2 >= $uidx && $uidx > $p1 ) {
    $cr = ($uidx - $p1) * 2 ;
    $cg = 255 - (($uidx - $p1) / 3) ;
    $cb = 255 - (($uidx - $p1) * 2) ;
  }
  if( $uidx > $p2 ) {
    $cr = 255 ;
    $cg = 255 - ($uidx - $p3) ;
    $cb = 0 ;
  }

  return sprintf( "#%02x%02x%02x", $cr, $cg, $cb ) ;
}

function makeSpeedIndicatorSub( $spd, $spds, $spdx, $spdm, $unit, $cmax ) {
  $c = 255 ;
  if( $spd > $spds && $spd <= $spdx ) {
    $c = floor( ( $spd - $spds ) / $unit ) ;
    $c = ($c > $cmax) ? $cmax : $c ;
  }
  if( $spd > $spdx && $spd <= $spdm ) {
    $c = $cmax - floor( ( $spd - $spdx ) / $unit ) ;
    $c = ($c < 0 ) ? 0 : $c ;
  } 

  return $c ;
}

/* debug */
function debug() {
  global $rpos, $wData, $cData, $wDataMst, $cDataMst ;
  echo "pos: " ;
  print_r( $rpos ) ;
  echo "wData: " ;
  print_r( $wData ) ;
  echo "cData: " ;
  print_r( $cData ) ;
  echo "wDataMst: " ;
  print_r( $wDataMst ) ;
  echo "cDataMst: " ;
  print_r( $cDataMst ) ;
/*
  debugSub( "pos", $rpos ) ;
  echo "\n" ;
  debugSub( "wData", $wData ) ;
  echo "\n" ;
  debugSub( "wDataMst", $wDataMst ) ;
  echo "\n" ;
  debugSub( "cData", $cData ) ;
  echo "\n" ;
  debugSub( "cDataMst", $cDataMst ) ;
  echo "\n" ;
*/

  return ;
}

/* debug sub */
function debugSub( $name, $arr) {
  echo "$name{ " ;
  foreach( $arr as $k => $v ) {
    if( is_array($v) ) {
      debugSub($k, $v) ;
    } else {
      echo "$k=>$v," ;
    }
  }
  echo " }," ;
  return ;
}

/* microtime */
function microtime_float()
{
  list($usec, $sec) = explode(" ", microtime());
  return ((float)$usec + (float)$sec);
}
?>
